/**
 * The Paper Tape Project -- Punching Subproject
 * Windows NT User mode driver
 *
 * This Driver should work at Windows NT/2000/XP systems.
 * It uses the WinIo library (http://www.internals.com/)
 * for kernel mode calls.
 *
 * It compiles with MinGW/MSYS, C++!. No Microsoft development
 * environment or DDK toolkit neccessary.
 *
 *
 * Started at 02.09.2008 at winXP development workstation
 * in susanne@M66.
 *
 * (c) 2008 Sven Kppel
 *
 *
 **/

#include <stdio.h>
#include <stdlib.h>
#include "backend.h"

/* Stupid */
void puncher_backend_stop(PuncherBackend* c);

/* Windows/WinIo specific header fil1es */
#include <windows.h>
#include "backend.win/WinIo.h"

/* Parport positions */
#define PARPORT_BASE_REGISTER     0x378
#define PARPORT_DATA_REGISTER     PARPORT_BASE_REGISTER + 0
#define PARPORT_STATUS_REGISTER   PARPORT_BASE_REGISTER + 1
#define PARPORT_CONTROL_REGISTER  PARPORT_BASE_REGISTER + 2

/* Magic numbers. See <linux/include/linux/parport.h> at linux ;-) */
#define PARPORT_CONTROL_STROBE  0x01 /* C0 */
#define PARPORT_STATUS_BUSY    0x80 /* S7 */

/**
 * A windows workaround for the unix gettimeofday function which
 * gives the usecs back. 
 * 
 **/
//DWORD puncher_backend_mtime(DWORD since) {
	/*if(!since) {
		GetTickCount
	}*/

	/*long usec;
	union {
         long long ns100;
         FILETIME ft;
    } now;
     
    GetSystemTimeAsFileTime (&now.ft);
	
    usec = (long) ((now.ns100 / 10LL) % 1000000LL);
	return usec*1000 - since*/
//}

PuncherBackend* puncher_backend_new(int debug_flag) {
	PuncherBackend* c = (PuncherBackend*) malloc(sizeof(PuncherBackend));
	c->debug_flag = debug_flag;
	
	if( !InitializeWinIo() ) {
		printf("Bad error while initialisiing WinIo library!\n");
		return NULL;
	}
	
	PUNCHER_BACKEND_DPRINTF("WinIo initialized.\n");
	puncher_backend_stop(c);
	
	return c;
}

void puncher_backend_stop(PuncherBackend *c) {
	// turn Punch Instruction Off!
	if(! SetPortVal(PARPORT_CONTROL_REGISTER, PARPORT_CONTROL_STROBE, 1)) {
		printf("Backend Error: Could not turn PI off\n");
	}
	PUNCHER_BACKEND_DPRINTF("Switched PI off\n");
}

int puncher_backend_destroy(PuncherBackend* c) {
	puncher_backend_stop(c);
	// so that's it.
	ShutdownWinIo();
	PUNCHER_BACKEND_DPRINTF("WinIo shutdown finished. Backend destroyed.\n");
}

int puncher_backend_write_byte(PuncherBackend* c, unsigned char buf) {
	DWORD time, wait_time;
	unsigned char status;
	// so here we are. Let's punch! ;-)

	// Set Data pins
	if( !SetPortVal(PARPORT_DATA_REGISTER, buf, 1) )
		return -2;
	
	time = GetTickCount();
	
	Sleep(10 / 1000); // Take some time...


	/* anti "pulse" strobe => Turns Punch Instruction on */
	if( !SetPortVal(PARPORT_CONTROL_REGISTER, 0x0, 1) )
		return -3;

	PUNCHER_BACKEND_DPRINTF("data set; strobe pulsed...\n");
	
	// sleep the defined time
	Sleep(220 / 1000);
	
	// only for debugging, get status values:
	if( !GetPortVal(PARPORT_STATUS_REGISTER, (PDWORD) &status, 1))
		return -4;
	
	// end strobe + data pins
	PUNCHER_BACKEND_DPRINTF("Control is %x. Ending data pins... ", status);
	if( !SetPortVal(PARPORT_DATA_REGISTER, 0x0, 1) )
		return -5;
	PUNCHER_BACKEND_DPRINTF("+ strobe... ");
	if( !SetPortVal(PARPORT_STATUS_REGISTER, PARPORT_CONTROL_STROBE, 1) )
		return -5;

	// Wait until Puncher is ready to give out Punch Ready
	wait_time = 40 - (GetTickCount() - time);
	PUNCHER_BACKEND_DPRINTF("\nWaiting for Puncher Ready (%i msec)...\n", wait_time);
	Sleep(40/1000);
	//if(wait_time > 0)
	//	Sleep(wait_time / 1000);

	// Check status another time.
	if( !GetPortVal(PARPORT_STATUS_REGISTER, (PDWORD) &status, 1) )
		return -6;
	if( (status & PARPORT_STATUS_BUSY) == PARPORT_STATUS_BUSY) {
		PUNCHER_BACKEND_DPRINTF("Still busy (%x).\n", status);
		return 1;
	} else {
		PUNCHER_BACKEND_DPRINTF("Finished successfully (%x)\n", status);
		return 0;
	}
} // puncher_backend_write_byte
